package bar.stundenplan;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Collections;
import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 */
public final class SatWrapper {

    private int nClauses;
    private int nVariables;
    private final int varTrue;
    private final int varFalse;
    private final LinkedList<Integer> shuffled;

    // kann null sein
    private ProcessBuilder processBuilder; // process builder zum solver
    private Process process; // process zum solver
    private BufferedWriter writer; // zum solver daten senden
    private BufferedReader reader; // vom solver daten empfangen
    private boolean[] state; // das resultat der variablenbelegungen

    public SatWrapper(String pfad, int cpus, int hVariables, int hClauses) throws Exception {
        shuffled = new LinkedList<>();

        if (pfad != null) {
            processBuilder = new ProcessBuilder(pfad, "--threads=" + cpus);
            processBuilder.redirectErrorStream(true);

            process = processBuilder.start();

            InputStream processInputStream = process.getInputStream();
            OutputStream processOutputStream = process.getOutputStream();

            writer = new BufferedWriter(new OutputStreamWriter(processOutputStream));
            reader = new BufferedReader(new InputStreamReader(processInputStream));

            // write header
            writer.write("p cnf " + hVariables + " " + hClauses);
            writer.newLine();
        }

        varTrue = c_new_var();
        c_1(varTrue);
        varFalse = -varTrue;

    }

    int get_vars() {
        return nVariables;
    }

    int get_clauses() {
        return nClauses;
    }

    boolean is_true(int var) {
        if (var >= 0) {
            return state[var];
        }
        return !state[-var];
    }

    /**
     *
     * @return -1 (UNSAT), +1 (SAT) or 0 (UNKNOWN) <br>
     * @throws Exception
     */
    public int solve() throws Exception {
        int result = Config.SOLVER_UNKNOWN;

        if (writer != null) {
            // solver start initialisieren
            writer.flush();
            writer.close();

            while (process.isAlive() || reader.ready()) {
                String line = reader.readLine();

                if (line == null) {
                    continue;
                }

                if (line.startsWith("c ")) {
                    //System.out.println(line); // solver comments start with 'c '
                    continue;
                }

                if (line.startsWith("s SATISFIABLE")) {
                    result = Config.SOLVER_SAT;
                    state = new boolean[nVariables + 1];
                }

                if (line.startsWith("s UNSATISFIABLE")) {
                    result = Config.SOLVER_UNSAT;
                }

                if (line.startsWith("v ")) {
                    String[] split = line.split("\\s+");
                    for (int i = 1; i < split.length; i++) {
                        int var_no = Integer.parseInt(split[i]);
                        if (var_no > 0) {
                            state[var_no] = true;
                        }
                    }
                }

            }

            reader.close();

            //process.destroyForcibly();
            //System.out.println("Killed");
        }

        return result;
    }

    // CONSTRAINTS
    int c_new_var() {
        if (shuffled.isEmpty()) {
            for (int i = 0; i < 10000; i++) {
                nVariables++;
                if (Math.random() < 0.5) {
                    shuffled.addLast(nVariables);
                } else {
                    shuffled.addLast(-nVariables);
                }
            }
            Collections.shuffle(shuffled);
        }
        return shuffled.pollFirst();
    }

    int[] c_new_vars_array(int size) {
        int[] temp = new int[size];
        for (int i = 0; i < size; i++) {
            temp[i] = c_new_var();
        }
        return temp;
    }

    int[] c_new_vars_array_exact(int size, int amount) throws Exception {
        int[] temp = new int[size];
        LinkedList<Integer> list = new LinkedList<>();
        for (int i = 0; i < size; i++) {
            temp[i] = c_new_var();
            list.addLast(temp[i]);
        }
        SatWrapper.this.c_exactly_generic(list, amount);
        return temp;
    }

    int[][] c_new_vars_matrix(int rows, int cols) {
        int[][] temp = new int[rows][cols];
        for (int r = 0; r < rows; r++) {
            temp[r] = c_new_vars_array(cols);
        }
        return temp;
    }

    private void c_write(String line) throws Exception {
        nClauses++;
        if (writer != null) {
            writer.write(line);
            writer.newLine();
        }
    }

    void c_3(int a, int b, int c) throws Exception {
        c_write(a + " " + b + " " + c + " 0");
    }

    void c_2(int a, int b) throws Exception {
        c_write(a + " " + b + " 0");
    }

    void c_1(int a) throws Exception {
        c_write(a + " 0");
    }

    int c_at_most_one_tree(LinkedList<Integer> list) throws Exception {
        LinkedList<Integer> temp = new LinkedList<>(list);

        if (temp.isEmpty()) {
            temp.addLast(varFalse);
        }

        while (temp.size() >= 2) {
            int var1 = temp.pollFirst();
            int var2 = temp.pollFirst();
            int var3 = c_new_var_OR(var1, var2);
            c_not_both(var1, var2);
            temp.addLast(var3);
        }

        return temp.pollFirst();
    }

    void c_exactly_generic(LinkedList<Integer> list, int amount) throws Exception {
        if (amount == 0) {
            // all false
            for (int var : list) {
                c_1(-var);
            }
            return;
        }
        if (amount == list.size()) {
            // all true
            for (int var : list) {
                c_1(+var);
            }
            return;
        }
        if (amount > list.size()) {
            throw new Exception("c_exactly_generic(list, " + amount + ") amount > list.size() = " + list.size());
        }
        if (amount < 0) {
            throw new Exception("c_exactly_generic(list, " + amount + ") amount < 0");
        }
        if (amount == 1) {
            if (list.isEmpty()) {
                throw new Exception("c_exactly_generic(list, 1) called with empty list.");
            }
            int root = c_at_most_one_tree(list);
            c_1(root);
            return;
        }
        throw new Exception("Unknown amount: " + amount);
    }

    void c_create_pigeon(int rows, int cols) throws Exception {
        int[][] matrix = c_new_vars_matrix(rows, cols);

        for (int r = 0; r < rows; r++) {
            LinkedList<Integer> listR = new LinkedList<>();
            for (int c = 0; c < cols; c++) {
                listR.addLast(matrix[r][c]);
            }
            SatWrapper.this.c_exactly_generic(listR, 1);
            listR.clear();
        }

        for (int c = 0; c < cols; c++) {
            LinkedList<Integer> listC = new LinkedList<>();
            for (int r = 0; r < rows; r++) {
                listC.addLast(matrix[r][c]);
            }
            SatWrapper.this.c_exactly_generic(listC, 1);
            listC.clear();
        }

    }

    int c_new_var_OR(int var1, int var2) throws Exception {
        int var3 = c_new_var();
        c_2(-var1, +var3);
        c_2(-var2, +var3);
        c_3(+var1, +var2, -var3);
        return var3;
    }

    int c_new_var_AND(int var1, int var2) throws Exception {
        int var3 = c_new_var();
        c_2(+var1, -var3);
        c_2(+var2, -var3);
        c_3(-var1, -var2, +var3);
        return var3;
    }

    void c_not_both(int var1, int var2) throws Exception {
        c_2(-var1, -var2);
    }

    void c_exactly_generic(int[] a, int amount) throws Exception {
        LinkedList<Integer> list = new LinkedList<>();
        for (int i = 0; i < a.length; i++) {
            list.addLast(a[i]);
        }
        c_exactly_generic(list, amount);
    }

    void c_exactly_generic(int[][] a, int amount) throws Exception {
        LinkedList<Integer> list = new LinkedList<>();
        for (int i = 0; i < a.length; i++) {
            for (int j = 0; j < a[i].length; j++) {
                list.addLast(a[i][j]);
            }
        }
        c_exactly_generic(list, amount);
    }

    void c_fill_var(int[] a) {
        for (int i = 0; i < a.length; i++) {
            a[i] = c_new_var();
        }
    }

    void c_fill_var(int[][] a) {
        for (int i = 0; i < a.length; i++) {
            for (int j = 0; j < a[i].length; j++) {
                a[i][j] = c_new_var();
            }
        }
    }

    void c_fill_var(int[][][] a) {
        for (int i = 0; i < a.length; i++) {
            for (int j = 0; j < a[i].length; j++) {
                for (int k = 0; k < a[i][j].length; k++) {
                    a[i][j][k] = c_new_var();
                }
            }
        }
    }

    void c_equal(int[] a, int[] b) throws Exception {
        if (a.length != b.length) {
            throw new Exception("c_equal: Die Arrays sollten gleichlang sein!");
        }
        for (int i = 0; i < a.length; i++) {
            c_equal(a[i], b[i]);
        }
    }

    void c_equal(int a, int b) throws Exception {
        c_2(-a, +b);
        c_2(+a, -b);
    }

    int c_OR_array_rueckwaerts(int[] a, int start, int anzahl) throws Exception {
        LinkedList<Integer> list = new LinkedList<>();

        for (int i = 0; (i < anzahl) && (start - i >= 0); i++) {
            list.addLast(a[start - i]);
        }

        return c_OR_tree(list);
    }

    int c_OR_tree(LinkedList<Integer> list) throws Exception {
        LinkedList<Integer> temp = new LinkedList<>(list);

        if (temp.isEmpty()) {
            temp.addLast(varFalse);
        }

        while (temp.size() >= 2) {
            int x = temp.pollFirst();
            int y = temp.pollFirst();
            int z = c_new_var_OR(x, y);
            temp.addLast(z);
        }

        return temp.pollFirst();
    }

    int c_AND_tree(LinkedList<Integer> list) throws Exception {
        LinkedList<Integer> temp = new LinkedList<>(list);

        if (temp.isEmpty()) {
            temp.addLast(varFalse);
        }

        while (temp.size() >= 2) {
            int x = temp.pollFirst();
            int y = temp.pollFirst();
            int z = c_new_var_AND(x, y);
            temp.addLast(z);
        }

        return temp.pollFirst();
    }

    int c_FALSE() {
        return varFalse;
    }

    int c_TRUE() {
        return varTrue;
    }

    void c_at_most_generic(LinkedList<Integer> list, int maximum) throws Exception {
        LinkedList<Integer> temp = new LinkedList<>(list);
        if (temp.isEmpty()) {
            temp.addLast(varFalse);
        }
        // special case: trivial
        if (maximum >= temp.size()) {
            return;
        }
        // special case: all false
        if (maximum == 0) {
            for (int x : temp) {
                c_1(-x);
            }
            return;
        }
        // special case: at_most_one
        if (maximum == 1) {
            c_at_most_one_tree(temp);
            return;
        }
        // else choose...
        c_at_most_NETWORK(temp, maximum);
    }

    void c_at_least_generic(LinkedList<Integer> list, int minimum) throws Exception {
        list = new LinkedList<>(list);
        // special case: trivial
        if (minimum <= 0) {
            return;
        }
        // special case: all true
        if (minimum == list.size()) {
            for (int x : list) {
                c_1(+x);
            }
            return;
        }
        // special case: at_least_one
        if (minimum == 1) {
            int var = c_OR_tree(list);
            c_1(+var);
            return;
        }
        // else choose...
        c_at_least_NETWORK(list, minimum);
    }

    private void c_at_most_NETWORK(LinkedList<Integer> list, int maximum) throws Exception {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = maximum; i < list.size(); i++) {
            c_1(-list.get(i));
        }
    }

    private void c_at_least_NETWORK(LinkedList<Integer> list, int minimum) throws Exception {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = 0; i < minimum; i++) {
            c_1(+list.get(i));
        }
    }

    void c_bitonic_sort(LinkedList<Integer> list) throws Exception {
        c_fill_False_until_power_two(list);
        c_bitonic_sort_power_two(list);
    }

    private void c_fill_False_until_power_two(LinkedList<Integer> list) {
        int size = 1;
        while (size < list.size()) {
            size *= 2;
        }
        while (list.size() < size) {
            list.addLast(varFalse);
        }
    }

    private void c_bitonic_sort_power_two(LinkedList<Integer> list) throws Exception {
        for (int window = 2; window <= list.size(); window *= 2) {
            c_bitonic_sort_spiral(list, window);
            for (int difference = window / 2; difference >= 2; difference /= 2) {
                c_bitonic_sort_difference(list, difference);
            }
        }
    }

    private void c_bitonic_sort_spiral(LinkedList<Integer> list, int size) throws Exception {
        for (int i = 0; i < list.size(); i += size) {
            for (int i1 = i, i2 = i + size - 1; i1 < i2; i1++, i2--) {
                c_bitonic_comparator(list, i1, i2);
            }
        }
    }

    private void c_bitonic_sort_difference(LinkedList<Integer> list, int size) throws Exception {
        int half = size / 2;
        for (int i = 0; i < list.size(); i += size) {
            for (int j = 0; j < half; j++) {
                c_bitonic_comparator(list, i + j, i + j + half);
            }
        }
    }

    private void c_bitonic_comparator(LinkedList<Integer> result, int i1, int i2) throws Exception {
        if (i1 >= i2) {
            throw new Exception("c_bitonic_comparator: " + i1 + "," + i2 + " <-- ERROR!!!");
        }
        int a = result.get(i1);
        int b = result.get(i2);
        result.set(i1, c_new_var_OR(a, b));
        result.set(i2, c_new_var_AND(a, b));
    }

    void c_implication(int x, int y) throws Exception {
        c_2(-x, y);
    }

    void c_plan(String plan_data, int[][][] planWTS, int max_t, int max_s
            , boolean deaktiviere_spalte_min, boolean deaktiviere_spalte_max
            , boolean deaktiviere_zeile_min, boolean deaktiviere_zeile_max
            , boolean deaktiviere_zeile_min_ab, boolean deaktiviere_zeile_max_ab, boolean deaktiviere_zellen) throws Exception {
        String[] sSplit = plan_data.split(";");

        for (String line : sSplit) {
            String[] sSplit2 = line.split(",");
            if (sSplit2.length != 3) {
                continue;
            }
            int s = Integer.parseInt(sSplit2[0]);
            int t = Integer.parseInt(sSplit2[1]);
            String v = sSplit2[2];
            if ((s == -2) && (!deaktiviere_spalte_min)) {
                c_plan_tag_min(planWTS, max_s, t, v);
                continue;
            }
            if ((s == -1) && (!deaktiviere_spalte_max)) {
                c_plan_tag_max(planWTS, max_s, t, v);
                continue;
            }
            if (((s >= 0) && (t >= 0))&& (!deaktiviere_zellen)) {
                c_plan_stunde_tag(planWTS, s, t, max_s, max_t, v);
                continue;
            }

            if ((t == -1) && (!deaktiviere_zeile_min)) {
                c_plan_stunde_min(planWTS, s, max_t, v);
                continue;
            }
            if ((t == -2) && (!deaktiviere_zeile_max)) {
                c_plan_stunde_max(planWTS, s, max_t, v);
                continue;
            }

            if ((t == -3) && (!deaktiviere_zeile_max_ab)) {
                c_plan_stunde_ab_max(planWTS, s, max_s, max_t, v);
                continue;
            }
            if ((t == -4) && (!deaktiviere_zeile_min_ab)) {
                c_plan_stunde_ab_min(planWTS, s, max_s, max_t, v);
                continue;
            }
        }
    }

    private void c_plan_tag_min(int[][][] planWTS, int max_s, int t, String v) throws Exception {
        int min = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int s = 0; s < max_s; s++) {
                list.addLast(planWTS[w][t][s]);
            }
            c_at_least_generic(list, min);
        }
    }

    private void c_plan_tag_max(int[][][] planWTS, int max_s, int t, String v) throws Exception {
        int max = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int s = 0; s < max_s; s++) {
                list.addLast(planWTS[w][t][s]);
            }
            c_at_most_generic(list, max);
        }
    }

    private void c_plan_stunde_tag(int[][][] planWTS, int s, int t, int max_s, int max_t, String v) throws Exception {
        if ((s < max_s) && (t < max_t)) {
            if (Config.PLAN_P_P.equals(v)) {
                c_1(+planWTS[0][t][s]);
                c_1(+planWTS[1][t][s]);
            }
            if (Config.PLAN_P_M.equals(v)) {
                c_1(+planWTS[0][t][s]);
                c_1(-planWTS[1][t][s]);
            }
            if (Config.PLAN_M_P.equals(v)) {
                c_1(-planWTS[0][t][s]);
                c_1(+planWTS[1][t][s]);
            }
            if (Config.PLAN_M_M.equals(v)) {
                c_1(-planWTS[0][t][s]);
                c_1(-planWTS[1][t][s]);
            }
        }
    }

    private void c_plan_stunde_min(int[][][] planWTS, int s, int max_t, String v) throws Exception {
        int min = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int t = 0; t < max_t; t++) {
                list.addLast(planWTS[w][t][s]);
            }
            c_at_least_generic(list, min);
        }
    }

    private void c_plan_stunde_max(int[][][] planWTS, int s, int max_t, String v) throws Exception {
        int max = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int t = 0; t < max_t; t++) {
                list.addLast(planWTS[w][t][s]);
            }
            c_at_most_generic(list, max);
        }
    }

    private void c_plan_stunde_ab_max(int[][][] planWTS, int s_ab, int max_s, int max_t, String v) throws Exception {
        int max = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int t = 0; t < max_t; t++) {
                for (int s = s_ab; s < max_s; s++) {
                    list.addLast(planWTS[w][t][s]);
                }
            }
            c_at_most_generic(list, max);
        }

    }

    private void c_plan_stunde_ab_min(int[][][] planWTS, int s_ab, int max_s, int max_t, String v) throws Exception {
        int max = Integer.parseInt(v);
        for (int w = 0; w < 2; w++) {
            LinkedList<Integer> list = new LinkedList<>();
            for (int t = 0; t < max_t; t++) {
                for (int s = s_ab; s < max_s; s++) {
                    list.addLast(planWTS[w][t][s]);
                }
            }
            c_at_least_generic(list, max);
        }

    }

    void c_planV(String plan_data, LinkedList<Integer>[][][] planWTS_list, int max_t, int max_s) throws Exception {
        String[] sSplit = plan_data.split(";");

        int[][] matrixST = new int[max_s][max_t];

        // Spalten- und Zeilensummen zuerst
        for (String line : sSplit) {
            String[] sSplit2 = line.split(",");
            if (sSplit2.length != 3) {
                continue;
            }
            int s = Integer.parseInt(sSplit2[0]);
            int t = Integer.parseInt(sSplit2[1]);
            String v = sSplit2[2];
            // tag min
            if (s == -2) {
                c_planV_setze_tag_auf(matrixST, t, max_s, Integer.parseInt(v));
                continue;
            }
            // stunde min
            if (t == -1) {
                c_planV_setze_stunde_auf(matrixST, s, max_t, Integer.parseInt(v));
                continue;
            }
            // ab min
            if (t == -4) {
                c_planV_setze_stunde_ab_auf(matrixST, s, max_t, max_s, Integer.parseInt(v));
                continue;
            }
        }

        // Zellensummen überschreiben
        for (String line : sSplit) {
            String[] sSplit2 = line.split(",");
            if (sSplit2.length != 3) {
                continue;
            }
            int s = Integer.parseInt(sSplit2[0]);
            int t = Integer.parseInt(sSplit2[1]);
            String v = sSplit2[2];
            if ((s >= 0) && (t >= 0) && (s < max_s) && (t < max_t)) {
                matrixST[s][t] = Integer.parseInt(v);
                continue;
            }
        }

        // forciere jeweiliges Minimum
        for (int w = 0; w < 2; w++) {
            for (int s = 0; s < max_s; s++) {
                for (int t = 0; t < max_t; t++) {
                    c_at_least_generic(planWTS_list[w][t][s], matrixST[s][t]);
                }
            }
        }
    }

    private void c_planV_setze_tag_auf(int[][] matrixST, int t, int max_s, int wert) {
        for (int s = 0; s < max_s; s++) {
            matrixST[s][t] = Math.max(matrixST[s][t], wert);
        }
    }

    private void c_planV_setze_stunde_auf(int[][] matrixST, int s, int max_t, int wert) {
        for (int t = 0; t < max_t; t++) {
            matrixST[s][t] = Math.max(matrixST[s][t], wert);
        }
    }

    private void c_planV_setze_stunde_ab_auf(int[][] matrixST, int s_ab, int max_t, int max_s, int wert) {
        for (int s = s_ab; s < max_s; s++) {
            for (int t = 0; t < max_t; t++) {
                matrixST[s][t] = Math.max(matrixST[s][t], wert);
            }
        }
    }

    Process get_process() {
        return process;
    }

}
